package PVE::AuthCookieHandler;

use strict;

use Carp;
use CGI '3.12';
use mod_perl2 '1.9922';

use Apache::AuthCookie::Util;
use Apache2::RequestRec;
use Apache2::RequestUtil;
use Apache2::Log;
use Apache2::Access;
use Apache2::Response;
use Apache2::Util;
use APR::Table;
use Apache2::Const qw(:common M_GET HTTP_FORBIDDEN HTTP_MOVED_TEMPORARILY);
use URI::Escape;
use Digest::SHA1;
use PVE::SafeSyslog;
use Authen::PAM qw(:constants);
use PVE::Utils;
use base qw(Apache2::AuthCookie);
use Encode;

my $secret = (split (/\s/, `md5sum /etc/pve/pve-root-ca.key`))[0];
my $cookie_timeout = 2400; # seconds

my $valid_groups = PVE::Utils::roleinfo();

# password should be utf8 encoded
sub pam_is_valid_user {
    my ($username, $password) = @_;

    # user (www-data) need to be able to read /etc/passwd /etc/shadow

    my $pamh = new Authen::PAM ('common-auth', $username, sub {
	my @res;
	while(@_) {
	    my $msg_type = shift;
	    my $msg = shift;
	    push @res, (0, $password);
	}
	push @res, 0;
	return @res;
    });

    if (!ref ($pamh)) {
	my $err = $pamh->pam_strerror($pamh);
	die "Error during PAM init: $err";
    }

    my $res;

    if (($res = $pamh->pam_authenticate(0)) != PAM_SUCCESS) {
	my $err = $pamh->pam_strerror($res);
	die "auth failed: $err";
    }

    if (($res = $pamh->pam_acct_mgmt (0)) != PAM_SUCCESS) {
	my $err = $pamh->pam_strerror($res);
	die "auth failed: $err";
    }

    $pamh = 0; # call destructor

    return 1;
}

sub _is_valid_user {
    my ($username, $password) = @_;

    my $valid = 0;

    eval {
	$valid = pam_is_valid_user ($username, $password);
    };
    my $err = $@;

    if ($err) {
	syslog ('info', $err);
	return undef;
    }

    return undef if !$valid;

    my ($name, $passwd, $uid, $gid) = getpwnam ($username);
    my $groupname = getgrgid($gid) || 'nogroup';

    if (!$valid_groups->{$groupname}) {
	syslog ('info', "auth failed: group '$groupname' is not in the list of allowed groups");
	return undef;	
    }

    return $groupname, $username;
}

sub authen_cred {
    my $self = shift;
    my $r = shift;
    my ($username, $password) = @_;
    my $group;

    # login.pl encodes data in uft-8
 
    ($group, $username) = _is_valid_user ($username, $password);

    if ($group) {
	my $timestamp = time();
	my $session_key =
	    $username . '::' . $group . '::' . $timestamp . '::' . Digest::SHA1::sha1_hex($username, $group, $timestamp, $secret);
	return $session_key;
    }

    return undef;
}

sub authen_ses_key {
    my $self = shift;
    my $r = shift;
    my $session_key = shift;

    my ($username, $group, $time, $mac) = split /::/, $session_key;

    my $time_diff = time() - $time;

    my $uri = $r->uri;

    if (($time_diff < $cookie_timeout) && 
	(Digest::SHA1::sha1_hex($username, $group, $time, $secret) eq $mac )) {
	# update cookie
	my $timestamp = time();
	$session_key = $username . '::' . $group . '::' . $timestamp . '::' . 
	    Digest::SHA1::sha1_hex($username, $group, $timestamp, $secret);
	# update session key, but not for web services
	$self->send_cookie ($r, $session_key) if $uri !~ m|^/ws/| ;
	return $username;
    }

    return undef;
}

sub group {
    my ($self, $r, $args) = @_;

    my $cookie = $self->key ($r);
 
    my ($username, $group) = split /::/, $cookie;

    if ($args eq $group) {
	return OK;
    }

    return FORBIDDEN;
}

sub login_form {  
  my ($self, $r) = @_;

  my $uri = $r->uri;

  # no login for for web services
  if ($uri =~ m|^/ws/|) {
      my $code = AUTH_REQUIRED;
      $r->custom_response($code, '');
      return $code;
  }

  return $self->SUPER::login_form ($r);
}

1;
