package PVE::SOAPSerializer;

use strict;
use SOAP::Lite;
use vars qw(@ISA);

@ISA = qw (SOAP::Serializer);

sub new {
    my $class = shift;

    my $self = $class->SUPER::new (@_);
    
# SOAP Serializer bug fix:
# "a string with embeded URI 'http://exsample.com'" is encoded as URI!
# should be a string instead
#        'anyURI' => 
#	      [95, sub { $_[0] =~ /^(urn:)|(http:\/\/)/i; }, 'as_anyURI'],
# regex should be: /^((urn:)|(http:\/\/))/i;
# so we disbale that
    delete $self->{_typelookup}->{'anyURI'};

    return $self;
}

package PVE::ConfigServer;

use strict;
use vars qw(@ISA);
use Carp;
use PVE::SafeSyslog;
use File::stat;
use IO::File;
use Fcntl qw(:flock);
use MIME::Base64;
use PVE::Cluster;
use PVE::Utils;
use PVE::Config;
use IO::Socket::INET;
use Digest::SHA1;
use PVE::QemuServer;
use PVE::APLInfo;
use IPC::Open2;
use PVE::OpenVZ;
use PVE::Qemu;

use base 'Exporter';
our @EXPORT = qw($pve_config_daemon);
our $pve_config_daemon;

sub alive {
    my ($class) = @_;

    return 1;
}

sub ping {
    my ($class) = @_;

    my $cinfo = PVE::Cluster::clusterinfo ();

    my $status = { time => time (), insync => 1 };

    $status->{uptime} = PVE::Utils::get_uptime ();
    $status->{cpuinfo} = PVE::Utils::get_cpu_info ();
    $status->{meminfo} = PVE::Utils::get_memory_info ();   
    $status->{hdinfo}->{root} = PVE::Utils::get_hd_info ('/');   
    $status->{hdinfo}->{data} = PVE::Utils::get_hd_info ('/var/lib/vz');   

    my $procstat = PVE::Utils::read_proc_stat();
    $status->{cpu} = $procstat->{cpu};
    $status->{wait} = $procstat->{wait};

    my $syncstatus = PVE::Config::read_file ("syncstatus");

    foreach my $ni (@{$cinfo->{nodes}}) {
	my $cid = $ni->{cid};
	next if $cinfo->{local}->{cid} == $cid; # skip local CID
	my $lastsync = defined ($syncstatus->{$cid}) ? 
	    $syncstatus->{$cid}->{lastsync} : 0;
	$status->{"lastsync_$cid"} = $lastsync;
	my $sdiff = time() - $lastsync;
	$sdiff = 0 if $sdiff < 0;
	$status->{insync} = 0 if ($sdiff > (60*3));
    }

    return $status;
}

sub vzlist {
    my ($class) = @_;

    my $res = {};

    # openvz
    eval {
	$res = PVE::OpenVZ::vmlist();
    };

    my $err = $@;

    if ($err) {
	syslog ('err', "ERROR: $err");
    } else {
	$res->{openvz} = 1;
    }

    # qemu
    eval {

	my $qmlist = PVE::Qemu::vmlist();

	foreach my $vekey (keys %$qmlist) {
	    if (!$res->{$vekey}) {
		$res->{$vekey} = $qmlist->{$vekey};
	    } else {
		syslog ('err', "found duplicated ID '$vekey' - ignoring qemu instance\n");
	    }
	}
    };
    
    $err = $@;

    if ($err) {
	syslog ('err', "ERROR: $err");
    } else {
	$res->{qemu} = 1;
    }

    $res->{lasttime} = time();

    my $pc = PVE::Config::update_file ('pcounter', 'vzlist');
    $res->{version} = $pc->{vzlist};

    return $res;
}

sub vmlogview {
    my ($class, $cid, $veid, $service) = @_;

    my $filename = "/var/lib/vz/private/$veid/var/log/syslog";

    if ($service eq 'init') {
	$filename = "/var/lib/vz/private/$veid/var/log/init.log";
    } elsif ($service eq 'syslog') {
	# some systems (rh,centos) logs to messages instead
	my $msglog = "/var/lib/vz/private/$veid/var/log/messages";
	if ((! -f $filename) && (-f $msglog)) {
	    $filename = $msglog;
	}
    }

    my $lines = [];

    my $limit = 200;

    open (TMP, "tail -$limit $filename|");
    while (my $line = <TMP>) {
	chomp $line;
	push @$lines, $line;
    }
    close (TMP);

    return $lines;
}


# sub change_vmconfig {
#     my ($class, $veid, $type, $settings) = @_;

#     die "unknown virtualization type '$type'\n" if !($type eq 'openvz' || $type eq 'qemu');

#     my $res;

#     $res->{vzlist} = $class->vzlist();

#     if (my $d = $res->{vzlist}->{"VEID_$veid"}) {
# 	die "virtualization type mismatch" if $type ne $d->{type};

# 	if ($d->{type} eq 'openvz') {
	    
# 	    die "not implemented";
# 	    $res->{config} = PVE::Config::get_veconfig ($veid);

# 	} elsif ($d->{type} eq 'qemu') {

# 	    PVE::QemuServer::change_config ($veid, $settings);
# 	    $res->{config} = PVE::QemuServer::load_config ($veid);

# 	} else {
# 	    return undef;
# 	} 
# 	return $res;
#     } else {
# 	return undef;
#     }
# }

sub vmconfig {
    my ($class, $veid, $type) = @_;

    die "unknown virtualization type '$type'\n" if !($type eq 'openvz' || $type eq 'qemu');

    my $res;

    $res->{vzlist} = $class->vzlist();

    if (($type eq 'qemu') && !$res->{vzlist}->{qemu}) {
	die "unable to get qemu-server vm list - server not running?\n";
    }
    if (($type eq 'openvz') && !$res->{vzlist}->{openvz}) {
	die "unable to get openvz vm list?\n";
    }

    if (my $d = $res->{vzlist}->{"VEID_$veid"}) {
	die "virtualization type mismatch" if $type ne $d->{type};

	if ($d->{type} eq 'openvz') {
	    $res->{config} = PVE::Config::get_veconfig ($veid);
	} elsif ($d->{type} eq 'qemu') {
	    $res->{config} = PVE::Config::get_qmconfig  ($veid);
	} else {
	    return undef;
	} 
	return $res;
    } else {
	return undef;
    }
}

sub cluster_vzlist {
    my ($class, $cid, $vzlist) = @_;

    my $newlist = PVE::Config::update_file ('vzlist', $vzlist, $cid);

    my $vmops = PVE::Config::read_file ("vmops");

    PVE::Utils::foreach_vmrec ($vmops, sub {
	my ($cid, $vmid, $d, $ckey, $vmkey) = @_;
	my $old = $newlist->{$ckey}->{$vmkey};

	# command still running ?
	my $pstart;
	if ($old && PVE::Utils::check_process ($d->{pid}, $d->{pstart})) {
 
	    $old->{status} = $d->{command};

	    if ($d->{command} eq 'migrate') {
		PVE::Utils::foreach_vmrec ($newlist, sub {
		    my ($ncid, $nvmid, $nd) = @_;
		    $nd->{status} = 'migrate' if ($nvmid eq $vmid);
		});
	    }
	}
    });

    return $newlist;
}

# start long running workers
# $data append to the returned uniquely identifier, which
# has the following format: "UPID:$pid-$pstart:$startime:$dtype:$data"
# STDIN is redirected to /dev/null
# STDOUT,STDERR are redirected to the filename returned by upid_decode
# that file is locked wit flock to make sure only one process 
# is writing it

sub __fork_worker {
    my ($dtype, $data, $function) = @_;

    my $cpid;

    $dtype = 'unknown' if !defined ($dtype);

    $data = '' if !defined ($data);

    my $starttime = time ();

    my @psync = POSIX::pipe();

    # detect filename with faked PID
    my $tmp = PVE::Utils::upid_decode ("UPID:0-0:0:$dtype:$data"); 
    my $filename = $tmp->{filename};

    my $lockfh;
    # lock output file
    if ($filename) {

	$lockfh = IO::File->new ($filename, O_WRONLY|O_CREAT) ||
	    die "unable to open output file - $!\n";

	my $wwwid = getpwnam('www-data');	    
	chown $wwwid,  $filename;

	if (!flock ($lockfh, LOCK_EX|LOCK_NB)) {
	    undef $lockfh; # close
	    die "unable to lock output file\n";
	}

	if (!truncate ($lockfh, 0)) {
	    die "unable to truncate output file - $!\n";
	}
    }

    if (($cpid = fork()) == 0) {

	$SIG{INT} = $SIG{QUIT} = $SIG{TERM} = sub { die "received interrupt\n"; };

	$SIG{CHLD} = $SIG{PIPE} = 'DEFAULT';

	# set sess/process group - we want to be able to kill the
	# whole process group
	POSIX::setsid(); 

	POSIX::close ($psync[0]);

	PVE::Config::inotify_close();

	# we close the socket 
	my $httpd = $pve_config_daemon->{_daemon};
	$httpd->close();

	# same algorythm as used inside SA

	# STDIN = /dev/null
	my $fd = fileno (STDIN);
	close STDIN;
	POSIX::close(0) if $fd != 0;

	if (!open (STDIN, "</dev/null")) {
	    POSIX::_exit (1); 
	    kill ('KILL', $$); 
	}

	# redirect STDOUT
	$fd = fileno(STDOUT);
	close STDOUT;
	POSIX::close (1) if $fd != 1;

	if ($filename) {
	    if (!open (STDOUT, ">&", $lockfh)) {
		POSIX::_exit (1); 
		kill ('KILL', $$); 
	    }

	    STDOUT->autoflush (1);
	} else {
	    if (!open (STDOUT, ">/dev/null")) {
		POSIX::_exit (1); 
		kill ('KILL', $$); 
	    }
	}
      
	#  redirect STDERR to STDOUT
	$fd = fileno (STDERR);
	close STDERR;
	POSIX::close(2) if $fd != 2;

	if (!open (STDERR, ">&1")) {
	    POSIX::_exit (1); 
	      kill ('KILL', $$); 
	  }
	
	STDERR->autoflush (1);

	my $pstart = PVE::Utils::read_proc_starttime ($$) ||
	    die "unable to read process starttime";

	my $upid = PVE::Utils::upid_encode ({
	    pid => $$, pstart => $pstart, starttime => $starttime,
	    type => $dtype, data => $data });

	# sync with parent
	POSIX::write ($psync[1], $upid, length ($upid));
	POSIX::close ($psync[1]);

	&$function ($upid);

	die "should not be reached";
    }

    POSIX::close ($psync[1]);

    # sync with child (wait until child starts)
    my $upid = '';
    POSIX::read($psync[0], $upid, 4096);
    POSIX::close ($psync[0]);

    if ($lockfh) {
	undef $lockfh; # close
    }

    my $uh = PVE::Utils::upid_decode ($upid);
    if (!$uh || 
	!($uh->{pid} == $cpid && $uh->{starttime} == $starttime &&
	  $uh->{type} eq $dtype && $uh->{data} eq $data)) {
	syslog ('err', "got strange upid - $upid\n");
    }

    PVE::Utils::register_worker ($cpid);
    
    return $upid;
}

# UPID: unique worker process descriptor
#
# general format used by fork_worker is 
# UPID:$pid-$pstart:$start:$type:$data
#
# $pid    ... process id of worker
# $pstart ... process start time from /proc/pid/stat
# $start  ... time (epoch) when process started
# $type   ... string to identity format of $data
# $data   ... arbitrary text
#
# speicalized format we use is 
# UPID:$pid-$pstart:$start:vmops:$command:$cid:$veid
#
# $command    ... create, start, stop, destroy
# $cid,$veid  ... cluster identity of VE 
#
# Note: PIDs are recycled, so to test if a process is still running
# we use (PID,PSTART) pair.

sub vmcommand {
    my ($class, $userid, $command, $cid, $veid, $code) = @_;

    my $remip;
    my $remcmd = [];

    $userid = 'unknown' if !$userid;

    my $cinfo = PVE::Cluster::clusterinfo ();

    if ($cid != $cinfo->{local}->{cid}) {
	$remip = $cinfo->{"CID_$cid"}->{ip};
	# we force tty allocation in order to tranfer signals (kill)
	$remcmd = ['/usr/bin/ssh', '-t', '-t', '-n', '-o', 'BatchMode=yes', $remip];
    }

    my $realcmd = sub {
	my $upid = shift;

	print "$upid\n";

	my $res = -1;
       
	eval {
	    $res = &$code ($upid, $remip, $remcmd, $cinfo);

	    my $ni = $cinfo->{"CID_$cid"};
	    my $rcon = PVE::ConfigClient::connect ($ni->{configport});
	    if (my $vzlist = $rcon->vzlist()->result) {
		PVE::Config::update_file ('vzlist', $vzlist, $cid);
	    }
	};

	my $err = $@;

	if ($err) {
	    syslog ('err', $err);
	    print STDERR "\n$err";
	    exit (-1);
	} 

	print STDERR "\n"; # flush
	exit ($res);
    };

    if (my $uid = __fork_worker ('vmops', "$command:$cid:$veid:$userid", $realcmd)) {

	PVE::Config::update_file ("vmops", $uid);
	  
	return $uid;                 ;
    }

    return undef;
}

sub apl_start_download {
    my ($class, $userid, $aplname) = @_;

    $userid = 'unknown' if !$userid;

    my $pkglist = PVE::APLInfo::load_data();

    my $data;

    if (!$pkglist || !$aplname || !($data = $pkglist->{'all'}->{$aplname})) {
	syslog ('err', "download failed: no aplinfo for appliance '$aplname'");
	return;
    }

    my $realcmd = sub {
	my $upid = shift;

	print "$upid\n";

	my $tmp = "/tmp/apldownload-$$-tmp.dat";

	eval {
	    my $msg = "starting download: $aplname";
	    syslog ('info', $msg);
	    print STDERR "$msg\n";

	    my $src = $data->{location};
	    my $dest = "/var/lib/vz/template/cache/$aplname";

	    if (-f $dest) {
		my $md5 = (split (/\s/, `md5sum '$dest'`))[0];

		if ($md5 && (lc($md5) eq lc($data->{md5sum}))) {
		    $msg = "file already exists $md5 - no need to download";
		    syslog ('info', $msg);
		    print STDERR "$msg\n";
		    return;
		}
	    }

	    local %ENV;
	    my $pvecfg = PVE::Config::read_file('pvecfg'); 
	    if ($pvecfg && $pvecfg->{http_proxy}) {
		$ENV{http_proxy} = $pvecfg->{http_proxy};
	    }

	    my @cmd = ('/usr/bin/wget', '--progress=dot:mega', '-O', $tmp, $src);
	    if (system (@cmd) != 0) {
		die "download failed - $!\n";
	    }

	    my $md5 = (split (/\s/, `md5sum '$tmp'`))[0];
		
	    if (!$md5 || (lc($md5) ne lc($data->{md5sum}))) {
		die "wrong checksum: $md5 != $data->{md5sum}\n";
	    }

	    if (system ('mv', $tmp, $dest) != 0) {
		die "unable to save file - $!\n";
	    }
	};

	my $err = $@;

	unlink $tmp;

	if ($err) {
	    syslog ('err', $err);
	    print STDERR "\n\ndownload failed: $err";
	    exit (-1);
	} 

	syslog ('info', "download finished");
	print STDERR "download finished\n";

	exit (0);
    };

    if (my $uid = __fork_worker ('apldownload', "$userid:$aplname", $realcmd)) {
	return $uid; 
    }

    return undef;
}

sub vmconfig_set {
    my ($class, $userid, $cid, $veid, $type, $settings) = @_;

    die "unknown virtualization type '$type'\n" if !($type eq 'openvz' || $type eq 'qemu');

    my $cinfo = PVE::Cluster::clusterinfo ();

    my $remip;
    my $remcmd = [];

    if ($cid != $cinfo->{local}->{cid}) {
	$remip = $cinfo->{"CID_$cid"}->{ip};
	$remcmd = ['/usr/bin/ssh', '-n', '-o', 'BatchMode=yes', $remip];
    }

    return if !$settings;

    my $param;

    foreach my $key (keys %$settings) {
	my $v = $settings->{$key};
	next if !defined ($v);
	if (ref ($v) eq 'ARRAY') {
	    foreach my $v1 (@$v) {
		push @$param, "--$key", $remip ? PVE::Utils::shellquote ($v1) : $v1;
	    }
	} else {
	    push @$param, "--$key", $remip ?  PVE::Utils::shellquote ($v) : $v;
	}
    }

    return if scalar (@$param) == 0;

    $remip = 'localhost' if !$remip;

    syslog ('info', "apply settings to VM $veid on node $cid ($remip)");

    my @cmd;

    if ($type eq 'openvz') {	
	@cmd = (@$remcmd, '/usr/bin/pvectl', 'vzset', $veid, @$param);
    } else {
	@cmd = (@$remcmd, '/usr/sbin/qm', 'set', $veid, @$param);
    }

    if (system (@cmd) != 0) {
	my $cmdstr = join (' ', @cmd);
	my $msg = "unable to apply VM settings, command failed: $cmdstr\n";
	syslog ('err', $msg);
	die "$msg\n";
    }

    my $msg = "VM $veid settings applied";
    syslog ('info', $msg);
}

# set cdrom for qemu/kvm
sub vmconfig_setcdrom {
    my ($class, $userid, $cid, $veid, $settings) = @_;

    my $cinfo = PVE::Cluster::clusterinfo ();

    my $remip;
    my $remcmd = [];

    if ($cid != $cinfo->{local}->{cid}) {
	$remip = $cinfo->{"CID_$cid"}->{ip};
	$remcmd = ['/usr/bin/ssh', '-n', '-o', 'BatchMode=yes', $remip];
    }

    my $param;

    foreach my $key (keys %$settings) {
	my $v = $settings->{$key};
	next if !defined ($v);
	push @$param, "--$key",  $remip ?  PVE::Utils::shellquote ($v) : $v;
    }

    return if scalar (@$param) == 0;

    $remip = 'localhost' if !$remip;

    syslog ('info', "setting cdrom on VM $veid on node $cid ($remip)");

    my @cmd = (@$remcmd, '/usr/sbin/qm', 'cdrom', $veid, @$param);

    if (system (@cmd) != 0) {
	my $cmdstr = join (' ', @cmd);
	my $msg = "unable to set cdrom, command failed: $cmdstr\n";
	syslog ('err', $msg);
	die "$msg\n";
    }

    my $msg = "VM $veid set cdrom";
    syslog ('info', $msg);
}

# delete unused qemu/kvm disk images
sub qemu_unlink_disk {
    my ($class, $userid, $cid, $veid, $filename) = @_;

    my $cinfo = PVE::Cluster::clusterinfo ();

    my $remip;
    my $remcmd = [];

    if ($cid != $cinfo->{local}->{cid}) {
	$remip = $cinfo->{"CID_$cid"}->{ip};
	$remcmd = ['/usr/bin/ssh', '-n', '-o', 'BatchMode=yes', $remip];
    }

    $remip = 'localhost' if !$remip;

    syslog ('info', "delete image '$filename' on VM $veid on node $cid ($remip)");

    my @cmd = (@$remcmd, '/usr/sbin/qm', 'unlink', $veid, $filename);

    if (system (@cmd) != 0) {
	my $cmdstr = join (' ', @cmd);
	my $msg = "unable to delete image, command failed: $cmdstr\n";
	syslog ('err', $msg);
	die "$msg\n";
    }

    my $msg = "VM $veid image '$filename' successfuly deleted";
    syslog ('info', $msg);
}

sub vmcommand_create {
    my ($class, $userid, $cid, $veid, $type, $settings) = @_;

    die "unknown virtualization type '$type'\n" if !($type eq 'openvz' || $type eq 'qemu');

    return $class->vmcommand ($userid, 'create', $cid, $veid, sub {
	my ($upid, $remip, $remcmd, $cinfo) = @_;
	

	my @cmd;

	if ($type eq 'openvz') {
	    @cmd = (@$remcmd, '/usr/bin/pvectl', 'vzcreate', $veid);
	} else {
	    @cmd = (@$remcmd, '/usr/sbin/qm', 'create', $veid);
	}

	foreach my $key (keys %$settings) {
	    my $v = $settings->{$key};
	    next if !defined ($v);
	    if (ref ($v) eq 'ARRAY') {
		foreach my $v1 (@$v) {
		    push @cmd, "--$key", $remip ?  PVE::Utils::shellquote ($v1) : $v1;
		}
	    } else {
		push @cmd, "--$key", $remip ?  PVE::Utils::shellquote ($v) : $v;
	    }
	}

	$remip = 'localhost' if !$remip;

	syslog ('info', "creating new VM $veid on node $cid ($remip)");

	my $cmdstr = join (' ', @cmd);
	print "$cmdstr\n";

	if (system (@cmd) != 0) {

	    my $msg = "unable to apply VM settings - $!";
	    syslog ('err', $msg);
	    print "$msg\n";
	    return -1;
	}

	my $msg = "VM $veid created";
	syslog ('info', $msg);
	print "$msg\n";

	return 0;
    });
}

sub vmcommand_destroy {
    my ($class, $cid, $veid, $type, $userid) = @_;

    die "unknown virtualization type '$type'\n" if !($type eq 'openvz' || $type eq 'qemu');

    return $class->vmcommand ($userid, 'destroy', $cid, $veid, sub {
	my ($upid, $remip, $remcmd, $cinfo) = @_;

	$remip = 'localhost' if !$remip;

	syslog ('info', "destroying VM $veid on node $cid ($remip)");

	my @cmd;

	if ($type eq 'openvz') {
	    @cmd = (@$remcmd, '/usr/sbin/vzctl', 'destroy', $veid);
	} else {
	    @cmd = (@$remcmd, '/usr/sbin/qm', 'destroy', $veid);
	}

	my $cmdstr = join (' ', @cmd);

	print "$cmdstr\n";

	if (system (@cmd) != 0) {
	    my $msg = "VM $veid destroy failed - $!";
	    syslog ('err', $msg);
	    print "$msg\n";
	    return -1;
	}
	
	my $msg = "VM $veid destroyed";
	syslog ('info', $msg);
	print "$msg\n";

	return 0;
    });
}

sub vmcommand_stop {
    my ($class, $cid, $veid, $type, $userid, $force) = @_;

    die "unknown virtualization type '$type'\n" if !($type eq 'openvz' || $type eq 'qemu');

    return $class->vmcommand ($userid, 'stop', $cid, $veid, sub {
	my ($upid, $remip, $remcmd, $cinfo) = @_;

	$remip = 'localhost' if !$remip;

	syslog ('info', "stopping VM $veid on node $cid ($remip)");

	my @cmd;

	if ($type eq 'openvz') {
	    @cmd = (@$remcmd, '/usr/sbin/vzctl', 'stop', $veid);
	    push @cmd, '--fast' if $force;
	} else {
	    @cmd = (@$remcmd, '/usr/sbin/qm', $force ? 'stop' : 'shutdown', $veid);
	}

	my $cmdstr = join (' ', @cmd);

	print "$cmdstr\n";

	if (system (@cmd) != 0) {
	    my $msg = "VM $veid stop failed - $!";
	    syslog ('err', $msg);
	    print "$msg\n";
	    return -1;
	}
	
	my $msg = "VM $veid stopped";
	syslog ('info', $msg);
	print "$msg\n";

	return 0;
    });
}

sub vmcommand_umount {
    my ($class, $cid, $veid, $type, $userid) = @_;

    die "unknown virtualization type '$type'\n" if $type ne 'openvz';

    return $class->vmcommand ($userid, 'umount', $cid, $veid, sub {
	my ($upid, $remip, $remcmd, $cinfo) = @_;

	$remip = 'localhost' if !$remip;

	syslog ('info', "unmounting VM $veid on node $cid ($remip)");

	my @cmd;

	@cmd = (@$remcmd, '/usr/sbin/vzctl', 'umount', $veid);

	my $cmdstr = join (' ', @cmd);

	print "$cmdstr\n";

	if (system (@cmd) != 0) {
	    my $msg = "VM $veid umount failed - $!";
	    syslog ('err', $msg);
	    print "$msg\n";
	    return -1;
	}
	
	my $msg = "VM $veid unmounted";
	syslog ('info', $msg);
	print "$msg\n";

	return 0;
    });
}

sub vmcommand_start {
    my ($class, $cid, $veid, $type, $userid) = @_;

    die "unknown virtualization type '$type'\n" if !($type eq 'openvz' || $type eq 'qemu');

    return $class->vmcommand ($userid, 'start', $cid, $veid, sub {
	my ($upid, $remip, $remcmd, $cinfo) = @_;

	$remip = 'localhost' if !$remip;

	syslog ('info', "starting VM $veid on node $cid ($remip)");

	my @cmd;

	if ($type eq 'openvz') {
	    @cmd = (@$remcmd, '/usr/sbin/vzctl', 'start', $veid);
	} else {
	    @cmd = (@$remcmd, '/usr/sbin/qm', 'start', $veid);
	}

	my $cmdstr = join (' ', @cmd);

	print "$cmdstr\n";

	if (system (@cmd) != 0) {
	    my $msg = "VM $veid start failed - $!";
	    syslog ('err', $msg);
	    print "$msg\n";
	    return -1;
	}
	
	my $msg = "VM $veid started";
	syslog ('info', $msg);
	print "$msg\n";

	return 0;
    });
}

sub vmcommand_restart {
    my ($class, $cid, $veid, $type, $userid) = @_;

    die "unknown virtualization type '$type'\n" if !($type eq 'openvz' || $type eq 'qemu');

    return $class->vmcommand ($userid, 'restart', $cid, $veid, sub {
	my ($upid, $remip, $remcmd, $cinfo) = @_;

	$remip = 'localhost' if !$remip;

	syslog ('info', "restarting VM $veid on node $cid ($remip)");

	my @cmd;

	if ($type eq 'openvz') {	
	    @cmd = (@$remcmd, '/usr/sbin/vzctl', 'restart', $veid);
	} else {
	    @cmd = (@$remcmd, '/usr/sbin/qm', 'reset', $veid);
	}
	my $cmdstr = join (' ', @cmd);

	print "$cmdstr\n";

	if (system (@cmd) != 0) {
	    my $msg = "VM $veid restart failed - $!";
	    syslog ('err', $msg);
	    print "$msg\n";
	    return -1;
	}
	
	my $msg = "VM $veid restarted";
	syslog ('info', $msg);
	print "$msg\n";

	return 0;
    });
}

sub vmcommand_migrate {
    my ($class, $userid, $cid, $veid, $type, $target, $online) = @_;

    die "unknown virtualization type '$type'\n" if !($type eq 'openvz' || $type eq 'qemu');

    my $cinfo = PVE::Cluster::clusterinfo ();

    return $class->vmcommand ($userid, 'migrate', $cid, $veid, sub {
	my ($upid, $remip, $remcmd, $cinfo) = @_;

	$remip = 'localhost' if !$remip;

	my $targetip = $cinfo->{"CID_$target"}->{ip};

	syslog ('info', "migrating VM $veid from node $cid ($remip) to node $target ($targetip)");

	my @cmd;

	if ($type eq 'openvz') {
	    @cmd = (@$remcmd, '/usr/sbin/vzmigrate');
	    push @cmd, '--online' if $online;
	    push @cmd, $targetip;
	    push @cmd, $veid;
	} else {
	    @cmd = (@$remcmd, '/usr/sbin/qmigrate');
	    push @cmd, '--online' if $online;
	    push @cmd, $targetip;
	    push @cmd, $veid;
	}

	my $cmdstr = join (' ', @cmd);

	print "$cmdstr\n";

	if (system (@cmd) != 0) {
	    my $msg = "VM $veid migration failed - $!";
	    syslog ('err', $msg);
	    print "$msg\n";
	    return -1;
	}
	
	my $msg = "VM $veid migration done";
	syslog ('info', $msg);
	print "$msg\n";

	return 0;
    });
}


sub __next_vnc_port {

    for (my $p = 5900; $p < 6000; $p++) {

	my $sock = IO::Socket::INET->new (Listen => 5,
					  LocalAddr => 'localhost',
					  LocalPort => $p,
					  ReuseAddr => 1,
					  Proto     => 0);

	if ($sock) {
	    close ($sock);
	    return $p;
	}
    }

    die "unable to find free vnc port";
}

sub create_vnc_proxy {
    my ($class, $cid, $veid, $userid) = @_;

    my $remip;
    my $remcmd = [];

    $userid = 'unknown' if !$userid;

    my $cinfo = PVE::Cluster::clusterinfo ();

    if ($cid != $cinfo->{local}->{cid}) {
	$remip = $cinfo->{"CID_$cid"}->{ip};
	$remcmd = ['/usr/bin/ssh', '-T', '-o', 'BatchMode=yes', $remip];
    }

    my $port = __next_vnc_port ();
    # generate ticket, olny first 8 character used by vnc
    my $ticket = Digest::SHA1::sha1_base64 ($userid, rand(), time());

    my $timeout = 30; 

    my $realcmd = sub {
	my $upid = shift;

	syslog ('info', "starting vnc proxy $upid\n");

	my $qmcmd = [@$remcmd, "/usr/sbin/qm", 'vncproxy', $veid , $ticket];

	my $qmstr = join (' ', @$qmcmd);

	# also redirect stderr (else we get RFB protocol errors)
	my @cmd = ('/bin/nc', '-l', '-p', $port, '-w', $timeout, '-c', "$qmstr 2>/dev/null");

	my $cmdstr = join (' ', @cmd);
	syslog ('info', "CMD: $cmdstr");

	if (system (@cmd) != 0) {
	    my $msg = "VM $veid vnc proxy failed - $?";
	    syslog ('err', $msg);
	    exit (-1);
	}

	exit (0);
    };

    if (my $uid = __fork_worker ('vncproxy', "$cid:$veid:$userid:$port:$ticket", $realcmd)) {
	return { port => $port, ticket => $ticket};
    }

    return undef;
    
}

sub create_vnc_console {
    my ($class, $cid, $veid, $type, $userid, $status) = @_;

    my $remip;
    my $remcmd = [];

    $userid = 'unknown' if !$userid;

    my $cinfo = PVE::Cluster::clusterinfo ();

    if ($cid != $cinfo->{local}->{cid}) {
	$remip = $cinfo->{"CID_$cid"}->{ip};
	$remcmd = ['/usr/bin/ssh', '-t', $remip];
    }

    my $port = __next_vnc_port ();
    # generate ticket, olny first 8 character used by vnc
    my $ticket = Digest::SHA1::sha1_base64 ($userid, rand(), time());

    my $timeout = 1; # immediately exit when last client disconnects

    my $realcmd = sub {
	my $upid = shift;

	syslog ('info', "starting vnc console $upid\n");

	# fixme: use ssl

	my $pwfile = "/tmp/.vncpwfile.$$";

	my $vzcmd;

	if ($type eq 'openvz') {
	    if ($status eq 'running') {
		$vzcmd = [ '/usr/sbin/vzctl', 'enter', $veid ];
	    } elsif ($status eq 'mounted') {
		$vzcmd = [ "/usr/bin/pvebash", $veid, 'root'];
	    } else {
		$vzcmd = [ "/usr/bin/pvebash", $veid, 'private'];
	    }
	} elsif ($type eq 'qemu') {
	    $vzcmd = [ "/usr/sbin/qm", 'monitor', $veid ];
	} else {
	    $vzcmd = [ '/bin/true' ]; # should not be reached
	}

	my @cmd = ('/usr/bin/vncterm', '-rfbport', $port, 
		   '-passwdfile', "rm:$pwfile",
		   '-timeout', $timeout, '-c', @$remcmd, @$vzcmd);

	my $cmdstr = join (' ', @cmd);
	syslog ('info', "CMD: $cmdstr");

	my $fh = IO::File->new ($pwfile, "w", 0600);
	print $fh "$ticket\n";
	$fh->close;

	if (system (@cmd) != 0) {
	    my $msg = "VM $veid console viewer failed - $?";
	    syslog ('err', $msg);
	    exit (-1);
	}

	exit (0);
    };

    if (my $uid = __fork_worker ('vncview', "$cid:$veid:$userid:$port:$ticket", $realcmd)) {
	
	#PVE::Config::update_file ("vncview", $uid);
	  
	return { port => $port, ticket => $ticket};
    }

    return undef;
    
}

sub service_cmd {
    my ($class, $service, $cmd) = @_;

    my $initd_cmd;

    ($cmd eq 'start' || $cmd eq 'stop' || $cmd eq 'restart' || $cmd eq 'reload' || $cmd eq 'awaken') ||
	die "unknown service command '$cmd': ERROR";
	
    if ($service eq 'postfix') {
	$initd_cmd = '/etc/init.d/postfix';
    } elsif ($service eq 'pvetunnel') {
	$initd_cmd = '/etc/init.d/pvetunnel';
    } elsif ($service eq 'pvemirror') {
	$initd_cmd = '/etc/init.d/pvemirror';
    } elsif  ($service eq 'apache') {
	if ($cmd eq 'restart') {
	    
	    $initd_cmd = '/usr/sbin/apache2ctl';
	    $cmd = 'graceful';
	} else {
	    die "invalid service cmd 'apache $cmd': ERROR";
	}
    } elsif  ($service eq 'network') {
	if ($cmd eq 'restart') {
	    return system ('(sleep 1; /etc/init.d/networking restart; /etc/init.d/postfix restart; /usr/sbin/apache2ctl graceful)&');
	} 
	die "invalid service cmd 'network $cmd': ERROR";
   } elsif  ($service eq 'ntpd') {
	# debian start/stop scripts does not work for us
	if ($cmd eq 'stop') {
	    system ('/etc/init.d/ntp stop');
	    #system ('/usr/bin/killall /usr/sbin/ntpd'); 
	} elsif ($cmd eq 'start') {
	    system ('/etc/init.d/ntp start');
	    system ('/sbin/hwclock --systohc');
	} elsif ($cmd eq 'restart') {
	    system ('/etc/init.d/ntp restart');
	    system ('/sbin/hwclock --systohc');
	    # restart cron/syslog to get right schedules and log time/dates
	    system ('/etc/init.d/sysklogd restart');
	    system ('/etc/init.d/cron restart');
	}
	return 0;
    } elsif  ($service eq 'syslog') {
	$initd_cmd = '/etc/init.d/sysklogd';
    } elsif  ($service eq 'cron') {
	$initd_cmd = '/etc/init.d/cron';
    } elsif  ($service eq 'sshd') {
	$initd_cmd = '/etc/init.d/ssh';
    } else {
	die "unknown service '$service': ERROR";
    }    

    my $servicecmd = "$initd_cmd $cmd";

    eval {
	my $res = PVE::Utils::run_command ([$initd_cmd, $cmd]);
        syslog ('info', $res) if $res;
        syslog ('info', "service command '$servicecmd' successful");
    };

    if (my $err = $@) {
	syslog ('err', "service command '$servicecmd' failed : $err");
    }
}

sub service_state {
    my ($class, $service) = @_;

    my $pid_file;

    if ($service eq 'postfix') {
	$pid_file = '/var/spool/postfix/pid/master.pid';
    } elsif  ($service eq 'apache') {
	$pid_file = '/var/run/apache2.pid';
    } elsif  ($service eq 'bind') {
	$pid_file = '/var/run/bind/run/named.pid';
    } elsif  ($service eq 'pvetunnel') {
	$pid_file = '/var/run/pvetunnel.pid';
    } elsif  ($service eq 'pvemirror') {
	$pid_file = '/var/run/pvemirror.pid';
    } elsif  ($service eq 'ntpd') {
	$pid_file = '/var/run/ntpd.pid';
    } elsif  ($service eq 'sshd') {
	$pid_file = '/var/run/sshd.pid';
    } else {
	die "unknown service '$service': ERROR";
    }    

    my $pid;
    if (my $fh = IO::File->new ($pid_file, "r")) {
	my $line = <$fh>;
	chomp $line;
	    
	if ($line  && ($line =~ m/^\s*(\d+)\s*$/)) {
	    $pid = $1;
	}
    }

    return 'running' if ($pid && kill (0, $pid));
    return 'stopped';
}

my $service_list = {
    apache => { short => 'WWW', long => 'Web Server' },
    pvetunnel => { short => 'ClusterTunnel', 
		    long => 'PVE Cluster Tunnel Daemon' },
    pvemirror => { short => 'ClusterSync', 
		    long => 'PVE Cluster Synchronization Daemon' },
    postfix => { short => 'SMTP', long => 'Simple Mail Tranfer Protocol' },
    ntpd => { short => 'NTP', long => 'Network Time Protocol' },
    sshd => { short => 'SSH', long => 'Secure Shell Daemon' },
    # bind => { short => 'BIND', long => 'Local DNS Cache' },
};

sub service_state_all {
    my ($class) = @_;

    my $res = {};

    foreach my $s (keys %{$service_list}) {
	$res->{$s} = $service_list->{$s};
	$res->{$s}->{status} = $class->service_state ($s);
    }

    return $res;
}

sub service_wait_stopped {
    my ($class, $timeout, @services) = @_;
    
    my $starttime = time();

    eval {
	while (1) {
	    my $wait = 0;
 
	    foreach my $s (@services) {
		if ($class->service_state ($s) eq 'running') {

		    if ((time() - $starttime) > $timeout) {
			syslog ('err', "unable to stop services (got timeout)");
			$wait = 0;
			last;
		    }

		    $class->service_cmd ($s, 'stop');
		    $wait = 1;
		}
	    }

	    if ($wait) {
		sleep (1);
	    } else {
		last;
	    }
	}
    };
}

sub restart_server {
    my ($class, $poweroff) = @_;

    if ($poweroff) {
	system ("(sleep 2;/sbin/poweroff)&");
    } else {
	system ("(sleep 2;shutdown -r now)&");
    }
}

sub check_worker {
    my ($class, $upid, $killit) = @_;

    if (my $upid_hash = PVE::Utils::upid_decode ($upid)) {

	my $pid = $upid_hash->{pid};

	# test if still running
	return 0 if !PVE::Utils::check_process ($pid, $upid_hash->{pstart});

	if ($killit) {

	    # send kill to process group (negative pid)
	    my $kpid = -$pid;

	    kill (15, $kpid); # send TERM signal

	    # give max 5 seconds to shut down
	    # note: waitpid only work for child processes, but not
	    # for processes spanned by other processes, so we use 
	    # kill to detect if the worker is still running
	    for (my $i = 0; $i < 5; $i++) {
		last if !kill (0, $kpid);
		sleep (1);
	    }
       
	    if (kill (0, $kpid)) {
		kill (9, $kpid); # kill if still alive
	    }

	    return 0; # killed, not running
	} else {
	    return 1; # running
	}
    }

    return 0;
}

sub ssh_authorized_keys {
    my ($class, $keys) = @_;

    my $fn =  '/root/.ssh/authorized_keys';
    my $fh;

    if ($keys) {
	-d '/root/.ssh' || mkdir '/root/.ssh', 0700;
	$fh = PVE::AtomicFile->open ($fn, "w") ||
	    die "cant open '$fn' - $! :ERROR";
	chomp $keys;
	print $fh "$keys\n";
	$fh->close (1);
    } else {
	if ($fh = IO::File->new ($fn, "r")) {
	    while (my $k = <$fh>) {
		$keys .= $k;
	    }
	}
    }

    return $keys;
}

sub kvm_version {
    my ($class) = @_;

    return PVE::QemuServer::kvm_version();
}

sub install_template {
    my ($class, $tmpname, $filename) = @_;

    PVE::Config::install_template ($tmpname, $filename);
}

sub delete_template {
    my ($class, $type, $filename) = @_;

    PVE::Config::delete_template ($type, $filename);
}

sub get_config_data {
    my ($self, $id, $full) = @_;

    return PVE::Config::read_file ($id, $full);
}

sub set_config_data {
    my ($self, $id, $data, $full) = @_;

    return PVE::Config::write_file ($id, $data, $full);
}

sub update_config_data {
    my ($self, $id, $data, @param) = @_;

    return PVE::Config::update_file ($id, $data, @param);
}

sub discard_config_changes {
    my ($self, $id, $full) = @_;

    return PVE::Config::discard_changes ($id, $full);
}

sub modify_user {
    my ($self, $username, $group, $pw, $comment) = @_;

    return PVE::Utils::modify_user ($username, $group, $pw, $comment);
}


package PVE::SOAPTransport;

use strict;
use vars qw(@ISA);
use SOAP::Transport::HTTP;
use PVE::SafeSyslog;
use POSIX qw(EINTR);
use POSIX ":sys_wait_h";
use IO::Handle;
use IO::Select;
use vars qw(@ISA);

# This is a quite simple pre-fork server

@ISA = qw(SOAP::Transport::HTTP::Daemon);

my $workers = {};

my $max_workers = 2;    # pre-forked worker processes
my $max_requests = 500; # max requests per worker

sub worker_finished {
    my $cpid = shift;

    syslog ('info', "worker $cpid finished");
}
    
sub finish_workers {
    local $!; local $?;    
    foreach my $cpid (keys %$workers) {
        my $waitpid = waitpid ($cpid, WNOHANG);
        if (defined($waitpid) && ($waitpid == $cpid)) {
            delete ($workers->{$cpid});
	    worker_finished ($cpid);
	}
    }
}

sub test_workers {
    foreach my $cpid (keys %$workers) {
	if (!kill(0, $cpid)) {
	    waitpid($cpid, POSIX::WNOHANG());
	    delete $workers->{$cpid};
	    worker_finished ($cpid);
	}
    }
}

sub start_workers {
    my $self = shift;

    my $count = 0;
    foreach my $cpid (keys %$workers) {
	$count++;
    }

    my $need = $max_workers - $count;

    return if $need <= 0;

    syslog ('info', "starting $need worker(s)");

    while ($need > 0) {
	my $pid = fork;

	if (!defined ($pid)) {
	    syslog ('err', "can't fork worker");
	    sleep (1);
	} elsif ($pid) { #parent
	    $workers->{$pid} = 1;
	    $0 = 'pvedaemon worker';
	    syslog ('info', "worker $pid started");
	    $need--;
	} else {
	    $SIG{TERM} = $SIG{QUIT} = 'DEFAULT';

	    eval {
		# try to init inotify
		PVE::Config::inotify_init();
	    };
	    syslog ('err', $@) if $@;


	    $self->handle_requests ();
	    exit (0);
	}
    }
}

sub terminate_server {

    foreach my $cpid (keys %$workers) {
	kill (15, $cpid); # TERM childs
    }

    # nicely shutdown childs (give them max 10 seconds to shut down)
    my $previous_alarm = alarm (10);
    eval {
	local $SIG{ALRM} = sub { die "Timed Out!\n" };
	
	1 while ((my $pid = waitpid (-1, 0)) > 0);

    };
    alarm ($previous_alarm);

    foreach my $cpid (keys %$workers) {
	!kill (0, $cpid) || kill (9, $cpid); # KILL childs still alive!
    }
}

sub handle {
    my $self = shift;
    my $daemon = $self->new;

    $self->{httpdaemon} = $daemon;
 
    eval {
	my $old_sig_chld = $SIG{CHLD};
	local $SIG{CHLD} = sub {
	    finish_workers ();
	    &$old_sig_chld(@_);
	};

	my $old_sig_term = $SIG{TERM};
	local $SIG{TERM} = sub { 
	    terminate_server ();
	    &$old_sig_term(@_);
	};
	local $SIG{QUIT} = sub { 
	    terminate_server();
	    &$old_sig_term(@_);
	};

	for (;;) { # forever
	    $self->start_workers ();
	    sleep (5); 
	    $self->test_workers ();
	}
    };
    my $err = $@;

    if ($err) {
	syslog ('err', "ERROR: $err");
    }
}

sub handle_requests {
    my $self = shift;

    my $daemon = $self->{httpdaemon};

    my $rcount = 0;


    my $sel = IO::Select->new();
    $sel->add ($daemon->{_daemon});

    my $timeout = 5;
    my @ready;
    while (1) {
	if (scalar (@ready = $sel->can_read($timeout))) {

	    my $c;
	    while (($c = $daemon->accept) || ($! == EINTR)) {

		next if !$c; # EINTR
	    
		# handle requests 
		while (my $r = $c->get_request) {
		    $daemon->request($r);
		    $daemon->SOAP::Transport::HTTP::Server::handle;
		    $c->send_response($daemon->response);
		}
		UNIVERSAL::can($c, 'shutdown') ? $c->shutdown(2) : $c->close(); 
		$c->close;

		$rcount++;

		last;
	    }

	    last if !$c || ($rcount >= $max_requests);

	} else {
	    # timeout
	    PVE::Config::poll(); # read inotify events
	}
    }
}

package PVE::ConfigClient;

use SOAP::Lite;

sub connect {
    my ($port) = @_;

    $port = 83 if !$port;
    
    my $timeout = ($port == 83) ? 120 : 10;

    my $soap = SOAP::Lite
	-> serializer (PVE::SOAPSerializer->new)
	-> uri('http://localhost/PVE/ConfigServer')
	-> on_fault (sub { 
	    my($soap, $res) = @_; 
	    die ref $res ? $res->faultstring : $soap->transport->status, "\n";
	})
	-> proxy("http://localhost:$port", timeout => $timeout);
    
    return $soap;
}

sub cluster_connect {
    my ($cinfo, $cid, $veid) = @_;

    if (!defined ($cid) || $cid !~ m/^\d+$/) {
	die "invalid cluster ID '$cid'";
    }
    if (!$veid || $veid !~ m/^\d+$/) {
	die "invalid VEID '$veid'";
    }

    my $ni = $cinfo->{"CID_$cid"} ||
	die "unable to determine cluster node '$cid'";

    return PVE::ConfigClient::connect ($ni->{configport});
}

1;
